from .data_source_postgresql import PostgreSQLDataSourceHelper, PostgreSQLDataSourceParams
from .storage_postgresql import StoragePostgreSQL, TopicDataStoragePostgreSQL
from .storage_postgresql_configuration import StoragePostgreSQLConfiguration
from .script_builder_postgresql import ScriptBuilderPostgreSQL



from typing import List

from watchmen_model.admin import Topic
from watchmen_storage import as_table_name
from .table_creator import build_unique_indexes_script, build_indexes_script, \
	build_columns_script, build_table_script
from watchmen_storage_rds import ScriptBuilder

from json import dumps
from datetime import datetime, date

from watchmen_utilities import ArrayHelper


class ScriptBuilderPostgreSQL(ScriptBuilder):
	
	def sql_insert(self, table_name: str, data: dict) -> str:
		columns, values = data.keys(), data.values()
		# noinspection PyTypeChecker
		script = f'''
INSERT INTO {table_name} ({", ".join(columns)})
VALUES ({insert_into_statement_build_values(values)});\n
'''
		return script
	
	def sql_update(self, table_name: str, primary_key: str, data: dict) -> str:
		sets = []
		for key, value in data.items():
			if key == primary_key:
				id_ = escape_value(value)
			set_script = f'{key} = {escape_value(value)}'
			sets.append(set_script)
		script = f'''
UPDATE {table_name} SET {", ".join(sets)} WHERE {primary_key} = {id_};\n'''
		return script
	
	def sql_create_table(self, topic: Topic) -> str:
		return f'{build_table_script(topic)}\n{create_index_for_table(as_table_name(topic))}'

	def sql_alert_table(self, topic: Topic, origin_topic: Topic) -> List[str]:
		return build_columns_script(topic, origin_topic)
	
	def sql_unique_indexes(self, topic: Topic) -> List[str]:
		return build_unique_indexes_script(topic)
	
	def sql_index(self, topic: Topic) -> List[str]:
		return build_indexes_script(topic)


def insert_into_statement_build_values(values: List) -> str:
	return ", ".join(ArrayHelper(values).map(escape_value).to_list())


def escape_value(value) -> str:
	if value:
		if isinstance(value, str):
			return f"'{value}'"
		elif isinstance(value, int):
			return f"{str(value)}"
		elif isinstance(value, dict):
			return f"'{dumps(value)}'"
		elif isinstance(value, list):
			return f"'{dumps(value)}'"
		elif isinstance(value, datetime):
			return f"""
'{value.strftime("%Y-%m-%d %H:%M:%S")}'
"""
		elif isinstance(value, date):
			return f"""
'{value.strftime("%Y-%m-%d")}'
"""
	else:
		return f"null"


def create_index_for_table(entity_name: str) -> str:
	return "\n".join([f'CREATE INDEX i_{entity_name}_tenant_id_ ON {entity_name} (tenant_id_)',
	                  f'CREATE INDEX i_{entity_name}_insert_time_ ON {entity_name} (insert_time_)',
	                  f'CREATE INDEX i_{entity_name}_update_time_ ON {entity_name} (update_time_)'])



from typing import Callable, Dict, List, Optional, Tuple, Union

from watchmen_model.admin import Factor, FactorType, is_aggregation_topic, is_raw_topic, Topic
from watchmen_storage import as_table_name
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank


# noinspection DuplicatedCode
def ask_column_name(factor: Factor) -> str:
	return factor.name.strip().lower().replace('.', '_').replace('-', '_').replace(' ', '_')


def varchar_column(precision: str) -> str:
	return f'VARCHAR({precision})'


def varchar_10(precision: Optional[str] = '10') -> str:
	return varchar_column('10' if is_blank(precision) else precision)


def varchar_20(precision: Optional[str] = '20') -> str:
	return varchar_column('20' if is_blank(precision) else precision)


def varchar_50(precision: Optional[str] = '50') -> str:
	return varchar_column('50' if is_blank(precision) else precision)


def varchar_100(precision: Optional[str] = '100') -> str:
	return varchar_column('100' if is_blank(precision) else precision)


def varchar_255(precision: Optional[str] = '255') -> str:
	return varchar_column('255' if is_blank(precision) else precision)


def decimal_column(precision: str) -> str:
	return f'DECIMAL({precision})'


def decimal_10_2(precision: Optional[str] = '10,2') -> str:
	return decimal_column('10,2' if is_blank(precision) else precision)


def decimal_32_6(precision: Optional[str] = '32,6') -> str:
	return decimal_column('32,6' if is_blank(precision) else precision)


FactorTypeMap: Dict[FactorType, Union[str, Callable[[Optional[str]], str]]] = {
	FactorType.SEQUENCE: 'DECIMAL(20)',

	FactorType.NUMBER: decimal_32_6,
	FactorType.UNSIGNED: decimal_32_6,

	FactorType.TEXT: varchar_255,

	# address
	FactorType.ADDRESS: 'VARCHAR(1024)',
	FactorType.CONTINENT: varchar_10,
	FactorType.REGION: varchar_10,
	FactorType.COUNTRY: varchar_10,
	FactorType.PROVINCE: varchar_10,
	FactorType.CITY: varchar_10,
	FactorType.DISTRICT: varchar_255,
	FactorType.ROAD: varchar_255,
	FactorType.COMMUNITY: varchar_100,
	FactorType.FLOOR: 'SMALLINT',
	FactorType.RESIDENCE_TYPE: varchar_10,
	FactorType.RESIDENTIAL_AREA: decimal_10_2,

	# contact electronic
	FactorType.EMAIL: varchar_100,
	FactorType.PHONE: varchar_50,
	FactorType.MOBILE: varchar_50,
	FactorType.FAX: varchar_50,

	# date time related
	FactorType.DATETIME: 'TIMESTAMP',
	FactorType.FULL_DATETIME: 'TIMESTAMP',
	FactorType.DATE: 'DATE',
	FactorType.TIME: 'TIME',
	FactorType.YEAR: 'SMALLINT',
	FactorType.HALF_YEAR: 'SMALLINT',
	FactorType.QUARTER: 'SMALLINT',
	FactorType.MONTH: 'SMALLINT',
	FactorType.HALF_MONTH: 'SMALLINT',
	FactorType.TEN_DAYS: 'SMALLINT',
	FactorType.WEEK_OF_YEAR: 'SMALLINT',
	FactorType.WEEK_OF_MONTH: 'SMALLINT',
	FactorType.HALF_WEEK: 'SMALLINT',
	FactorType.DAY_OF_MONTH: 'SMALLINT',
	FactorType.DAY_OF_WEEK: 'SMALLINT',
	FactorType.DAY_KIND: 'SMALLINT',
	FactorType.HOUR: 'SMALLINT',
	FactorType.HOUR_KIND: 'SMALLINT',
	FactorType.MINUTE: 'SMALLINT',
	FactorType.SECOND: 'SMALLINT',
	FactorType.MILLISECOND: 'SMALLINT',
	FactorType.AM_PM: 'SMALLINT',

	# individual
	FactorType.GENDER: varchar_10,
	FactorType.OCCUPATION: varchar_10,
	FactorType.DATE_OF_BIRTH: 'DATE',
	FactorType.AGE: 'SMALLINT',
	FactorType.ID_NO: varchar_50,
	FactorType.RELIGION: varchar_10,
	FactorType.NATIONALITY: varchar_10,

	# organization
	FactorType.BIZ_TRADE: varchar_10,
	FactorType.BIZ_SCALE: 'DECIMAL(9)',

	FactorType.BOOLEAN: 'SMALLINT',

	FactorType.ENUM: varchar_20,

	FactorType.OBJECT: 'JSON',
	FactorType.ARRAY: 'JSON'
}


# noinspection DuplicatedCode
def ask_column_type(factor: Factor) -> str:
	column_type = FactorTypeMap.get(factor.type)
	if isinstance(column_type, str):
		return column_type
	elif is_blank(factor.precision):
		return column_type()
	else:
		return column_type(factor.precision.strip())


def build_columns(topic: Topic) -> str:
	if is_raw_topic(topic):
		flatten_factors = ArrayHelper(topic.factors) \
			.filter(lambda x: x.flatten) \
			.map(lambda x: f'\t{ask_column_name(x)} {ask_column_type(x)},') \
			.to_list()
		if len(flatten_factors) == 0:
			return '\tdata_ JSON,'
		else:
			return '\n'.join(flatten_factors) + '\n\tdata_ JSON,'
	else:
		return ArrayHelper(topic.factors) \
			.filter(lambda x: '.' not in x.name) \
			.map(lambda x: f'\t{ask_column_name(x)} {ask_column_type(x)},') \
			.join('\n')


# noinspection DuplicatedCode
def build_aggregate_assist_column(topic: Topic) -> str:
	return f'\taggregate_assist_ JSON,' if is_aggregation_topic(topic) else ''


def build_version_column(topic: Topic) -> str:
	return f'\tversion_ DECIMAL(8),' if is_aggregation_topic(topic) else ''


# noinspection SqlResolve,DuplicatedCode
def build_columns_script(topic: Topic, original_topic: Topic) -> List[str]:
	entity_name = as_table_name(topic)
	original_factors: Dict[str, Factor] = ArrayHelper(original_topic.factors) \
		.to_map(lambda x: x.name.strip().lower(), lambda x: x)

	# noinspection SqlResolve
	def build_column_script(factor: Tuple[Factor, Optional[Factor]]) -> str:
		current_factor, original_factor = factor
		if original_factor is None:
			return f'ALTER TABLE {entity_name} ADD ({ask_column_name(factor[0])} {ask_column_type(factor[0])})'
		elif current_factor.flatten and not original_factor.flatten:
			return f'ALTER TABLE {entity_name} ADD ({ask_column_name(factor[0])} {ask_column_type(factor[0])})'
		else:
			return f'ALTER TABLE {entity_name} ALTER COLUMN ({ask_column_name(factor[0])} TYPE {ask_column_type(factor[0])})'

	if is_raw_topic(topic):
		factors = ArrayHelper(topic.factors) \
			.filter(lambda x: x.flatten) \
			.to_list()
	else:
		factors = topic.factors

	columns = ArrayHelper(factors) \
		.map(lambda x: (x, original_factors.get(x.name.strip().lower()))) \
		.map(build_column_script) \
		.to_list()

	if is_raw_topic(topic) and not is_raw_topic(original_topic):
		columns.append(f'ALTER TABLE {entity_name} ADD COLUMN data_ JSON')

	if is_aggregation_topic(topic) and not is_aggregation_topic(original_topic):
		columns.append(f'ALTER TABLE {entity_name} ADD COLUMN aggregate_assist_ JSON')
		columns.append(f'ALTER TABLE {entity_name} ADD COLUMN version_ DECIMAL(8)')

	return columns


def build_unique_indexes_script(topic: Topic) -> List[str]:
	index_groups: Dict[str, List[Factor]] = ArrayHelper(topic.factors) \
		.filter(lambda x: is_not_blank(x.indexGroup) and x.indexGroup.startswith('u-')) \
		.group_by(lambda x: x.indexGroup)

	# noinspection SqlResolve
	def build_unique_index(factors: List[Factor], index: int) -> str:
		return \
			f'CREATE UNIQUE INDEX u_{as_table_name(topic)}_{index + 1} ON as_table_name(topic) ' \
			f'({ArrayHelper(factors).map(lambda x: ask_column_name(x)).join(",")})'

	return ArrayHelper(list(index_groups.values())) \
		.map_with_index(lambda x, index: build_unique_index(x, index)).to_list()


def build_indexes_script(topic: Topic) -> List[str]:
	index_groups: Dict[str, List[Factor]] = ArrayHelper(topic.factors) \
		.filter(lambda x: is_not_blank(x.indexGroup) and x.indexGroup.startswith('i-')) \
		.group_by(lambda x: x.indexGroup)

	# noinspection SqlResolve
	def build_index(factors: List[Factor], index: int) -> str:
		return \
			f'CREATE INDEX u_{as_table_name(topic)}_{index + 1} ON {as_table_name(topic)} ' \
			f'({ArrayHelper(factors).map(lambda x: ask_column_name(x)).join(",")})'

	return ArrayHelper(list(index_groups.values())) \
		.map_with_index(lambda x, index: build_index(x, index)).to_list()


def build_table_script(topic: Topic) -> str:
	entity_name = as_table_name(topic)
	# noinspection SqlType
	return f'''
CREATE TABLE {entity_name} (
\tid_ DECIMAL(20),
{build_columns(topic)}
{build_aggregate_assist_column(topic)}
{build_version_column(topic)}
\ttenant_id_ VARCHAR(50),
\tinsert_time_ TIMESTAMP,
\tupdate_time_ TIMESTAMP,
\tCONSTRAINT pk_{entity_name} PRIMARY KEY (id_)
)'''



from typing import List, Optional

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

from watchmen_model.common import DataModel
from watchmen_model.system import DataSource, DataSourceParam
from watchmen_storage import DataSourceHelper
from watchmen_utilities import ArrayHelper, is_blank, is_not_blank, serialize_to_json
from .storage_postgresql import StoragePostgreSQL, TopicDataStoragePostgreSQL


def redress_url_by_psycopg2(url: str) -> str:
	if url.startswith('postgresql://'):
		return url.replace('postgresql://', 'postgresql+psycopg2://')
	else:
		return url


class PostgreSQLDataSourceParams(DataModel):
	echo: bool = False
	poolRecycle: int = 3600


class PostgreSQLDataSourceHelper(DataSourceHelper):
	def __init__(self, data_source: DataSource, params: PostgreSQLDataSourceParams = PostgreSQLDataSourceParams()):
		super().__init__(data_source)
		self.engine = self.acquire_engine(params)

	@staticmethod
	def acquire_engine_by_url(url: str, params: PostgreSQLDataSourceParams) -> Engine:
		return create_engine(
			redress_url_by_psycopg2(url),
			echo=params.echo,
			future=True,
			pool_recycle=params.poolRecycle,
			json_serializer=serialize_to_json,
			encoding='utf-8',
			supports_native_boolean=False
		)

	# noinspection DuplicatedCode
	@staticmethod
	def find_param(params: Optional[List[DataSourceParam]], key: str) -> Optional[str]:
		if params is None:
			return None

		for param in params:
			if is_not_blank(param.name) and param.name.strip().lower() == key:
				value = param.value
				if is_not_blank(value):
					return value.strip()
		return None

	@staticmethod
	def append_param(params: Optional[List[DataSourceParam]], key: str, value: str) -> List[DataSourceParam]:
		if params is None:
			params = []
		params.append(DataSourceParam(name=key, value=value))
		return params

	@staticmethod
	def build_url_search(params: Optional[List[DataSourceParam]]) -> str:
		if params is None:
			return ''
		else:
			return ArrayHelper(params).map(lambda param: f'{param.name}={param.value}').join('&')

	@staticmethod
	def acquire_engine_by_params(
			username: str, password: str, host: str, port: str, name: str,
			data_source_params: Optional[List[DataSourceParam]],
			params: PostgreSQLDataSourceParams
	) -> Engine:
		charset = PostgreSQLDataSourceHelper.find_param(data_source_params, 'client_encoding')
		if is_blank(charset):
			data_source_params = PostgreSQLDataSourceHelper.append_param(data_source_params, 'client_encoding', 'utf8')

		# please set search_path for current user who connect to db, make sure the default one is what you need.
		# otherwise, might incorrect schema used.
		# options = PostgreSQLDataSourceHelper.find_param(data_source_params, 'options')
		# if is_blank(options):
		# 	raise UnexpectedStorageException(
		# 		f'Parameter [options] must be specified for PostgreSQL data source [{name}].')
		# elif options.find('search_path') == -1:
		# 	raise UnexpectedStorageException(
		# 		f'Parameter [search_path] must be specified for PostgreSQL data source [{name}].')

		search = PostgreSQLDataSourceHelper.build_url_search(data_source_params)
		if is_not_blank(search):
			search = f'?{search}'
		url = f'postgresql+psycopg2://{username}:{password}@{host}:{port}/{name}{search}'
		return PostgreSQLDataSourceHelper.acquire_engine_by_url(url, params)

	def acquire_storage(self) -> StoragePostgreSQL:
		return StoragePostgreSQL(self.engine)

	def acquire_topic_data_storage(self) -> TopicDataStoragePostgreSQL:
		return TopicDataStoragePostgreSQL(self.engine)
#
#



from logging import getLogger
from typing import Any, Callable, List, Optional, Tuple

from sqlalchemy import Table, text

from watchmen_model.admin import FactorType, Topic
from watchmen_storage import as_table_name, EntityCriteria, EntitySort, Literal
from watchmen_storage_rds import build_sort_for_statement, SQLAlchemyStatement, \
	StorageRDS, TopicDataStorageRDS
from .table_creator import build_columns_script, build_indexes_script, \
	build_unique_indexes_script, build_table_script
from .where_build import build_criteria_for_statement, build_literal

# noinspection DuplicatedCode
logger = getLogger(__name__)


class StoragePostgreSQL(StorageRDS):
	def build_criteria_for_statement(
			self, tables: List[Table], statement: SQLAlchemyStatement,
			criteria: EntityCriteria,
			raise_exception_on_missed: bool = False) -> SQLAlchemyStatement:
		return build_criteria_for_statement(tables, statement, criteria, raise_exception_on_missed)

	def build_sort_for_statement(self, statement: SQLAlchemyStatement, sort: EntitySort) -> SQLAlchemyStatement:
		return build_sort_for_statement(statement, sort)


class TopicDataStoragePostgreSQL(StoragePostgreSQL, TopicDataStorageRDS):
	# noinspection SqlResolve,DuplicatedCode
	def create_topic_entity(self, topic: Topic) -> None:
		try:
			self.connect()
			entity_name = as_table_name(topic)
			# noinspection SqlType
			script = build_table_script(topic)
			self.connection.execute(text(script))
			# try to add index
			for unique_index_script in build_unique_indexes_script(topic):
				self.connection.execute(text(unique_index_script))
			for index_script in build_indexes_script(topic):
				self.connection.execute(text(index_script))
			self.connection.execute(
				text(f'CREATE INDEX i_{entity_name}_tenant_id_ ON {entity_name} (tenant_id_)'))
			self.connection.execute(
				text(f'CREATE INDEX i_{entity_name}_insert_time_ ON {entity_name} (insert_time_)'))
			self.connection.execute(
				text(f'CREATE INDEX i_{entity_name}_update_time_ ON {entity_name} (update_time_)'))
		except Exception as e:
			logger.error(e, exc_info=True, stack_info=True)
		finally:
			self.close()

	# noinspection DuplicatedCode
	def update_topic_entity(self, topic: Topic, original_topic: Topic) -> None:
		"""
		1. drop no column,\n
		2. factor indexes from original topic are dropped,\n
		3. factor indexes from topic are created,\n
		4. compatible column type changes are applied,\n
		5. any exception is ignored.
		"""
		try:
			self.connect()
			entity_name = as_table_name(topic)
			self.connection.execute(text(f"CALL DROP_INDEXES_ON_TOPIC_CHANGED('{entity_name}')"))
			# try to change column anyway, ignore when failed
			for column_script in build_columns_script(topic, original_topic):
				try:
					self.connection.execute(text(column_script))
				except Exception as e:
					logger.error(e, exc_info=True, stack_info=True)
			# try to add index
			for unique_index_script in build_unique_indexes_script(topic):
				try:
					self.connection.execute(text(unique_index_script))
				except Exception as e:
					logger.error(e, exc_info=True, stack_info=True)
			for index_script in build_indexes_script(topic):
				try:
					self.connection.execute(text(index_script))
				except Exception as e:
					logger.error(e, exc_info=True, stack_info=True)
			try:
				# noinspection SqlResolve
				self.connection.execute(
					text(f'CREATE INDEX i_{entity_name}_tenant_id_ ON {entity_name} (tenant_id_)'))
			except Exception as e:
				logger.error(e, exc_info=True, stack_info=True)
			try:
				# noinspection SqlResolve
				self.connection.execute(
					text(f'CREATE INDEX i_{entity_name}_insert_time_ ON {entity_name} (insert_time_)'))
			except Exception as e:
				logger.error(e, exc_info=True, stack_info=True)
			try:
				# noinspection SqlResolve
				self.connection.execute(
					text(f'CREATE INDEX i_{entity_name}_update_time_ ON {entity_name} (update_time_)'))
			except Exception as e:
				logger.error(e, exc_info=True, stack_info=True)
		except Exception as e:
			logger.error(e, exc_info=True, stack_info=True)
		finally:
			self.close()

	# noinspection SqlResolve
	def ask_synonym_columns_sql(self, table_name: str) -> str:
		return \
			f"SELECT TABLE_NAME, COLUMN_NAME, DATA_TYPE, " \
			f"CASE " \
			f"WHEN NUMERIC_PRECISION IS NOT NULL THEN DATA_TYPE || '(' || NUMERIC_PRECISION || ',' || NUMERIC_SCALE || ')' " \
			f"WHEN CHARACTER_MAXIMUM_LENGTH IS NOT NULL THEN DATA_TYPE || '(' || CHARACTER_MAXIMUM_LENGTH || ')' " \
			f"ELSE DATA_TYPE " \
			f"END AS COLUMN_TYPE, " \
			f"COLUMN_COMMENT " \
			f"FROM (SELECT PC.RELNAME AS TABLE_NAME, PA.ATTNAME AS COLUMN_NAME, PT.TYPNAME AS DATA_TYPE, " \
			f"INFORMATION_SCHEMA._PG_CHAR_MAX_LENGTH(INFORMATION_SCHEMA._PG_TRUETYPID(PA.*, PT.*), " \
			f"INFORMATION_SCHEMA._PG_TRUETYPMOD(PA.*, PT.*))   AS CHARACTER_MAXIMUM_LENGTH, " \
			f"INFORMATION_SCHEMA._PG_NUMERIC_PRECISION(INFORMATION_SCHEMA._PG_TRUETYPID(PA.*, PT.*), " \
			f"INFORMATION_SCHEMA._PG_TRUETYPMOD(PA.*, PT.*)) AS NUMERIC_PRECISION, " \
			f"INFORMATION_SCHEMA._PG_NUMERIC_SCALE(INFORMATION_SCHEMA._PG_TRUETYPID(PA.*, PT.*), " \
			f"INFORMATION_SCHEMA._PG_TRUETYPMOD(PA.*, PT.*)) AS NUMERIC_SCALE, " \
			f"COL_DESCRIPTION(PC.OID, PA.ATTNUM) AS COLUMN_COMMENT " \
			f"FROM PG_CLASS PC, PG_ATTRIBUTE PA, PG_TYPE PT " \
			f"WHERE UPPER(PC.RELNAME) = UPPER('{table_name}') " \
			f"AND PA.ATTRELID = PC.OID AND PA.ATTNUM > 0 AND PA.ATTTYPID = PT.OID " \
			f"ORDER BY PA.ATTNUM) AS T"

	def schema_column_data_type_to_factor_type(self, schema_column_data_type: str) -> Tuple[FactorType, Optional[str]]:
		if '(' in schema_column_data_type:
			index = schema_column_data_type.index('(')
			precision = schema_column_data_type[index + 1: len(schema_column_data_type) - 1]
			data_type = schema_column_data_type[: index].upper()
		else:
			precision = None
			data_type = schema_column_data_type.upper()

		if data_type in [
			'BIGINT', 'INT8', 'BIGSERIAL', 'SERIAL8', 'BIT', 'BIT VARYING', 'VARBIT', 'DOUBLE PRECISION', 'FLOAT8',
			'INTEGER', 'INT', 'INT4', 'MONEY', 'NUMERIC', 'DECIMAL', 'REAL', 'FLOAT4', 'SMALLINT', 'INT2',
			'SMALLSERIAL', 'SERIAL2', 'SERIAL', 'SERIAL4'
		]:
			return FactorType.NUMBER, precision
		elif data_type in ['TIMESTAMP', 'TIMESTAMP WITHOUT TIME ZONE', 'TIMESTAMP WITH TIME ZONE']:
			return FactorType.DATETIME, None
		elif data_type == 'DATE':
			return FactorType.DATE, None
		elif data_type in ['TIME', 'TIME WITHOUT TIME ZONE', 'TIME WITH TIME ZONE']:
			return FactorType.TIME, None
		elif data_type in [
			'CHARACTER', 'CHAR', 'CHARACTER VARYING', 'VARCHAR', 'TEXT'
		]:
			return FactorType.TEXT, precision
		elif data_type in ['JSON', 'JSONB']:
			return FactorType.OBJECT, None
		elif data_type in ['BOOLEAN', 'BOOL']:
			return FactorType.BOOLEAN, None
		else:
			return FactorType.TEXT, None

	# noinspection SqlResolve,SqlCaseVsIf
	def ask_synonym_indexes_sql(self, table_name: str) -> str:
		return \
			f"SELECT C.RELNAME AS TABLE_NAME, A.ATTNAME AS COLUMN_NAME, I.RELNAME AS INDEX_NAME, " \
			f"CASE WHEN X.INDISUNIQUE THEN 0 ELSE 1 END AS NON_UNIQUE " \
			f"FROM PG_INDEX X JOIN PG_CLASS C ON C.OID = X.INDRELID " \
			f"JOIN PG_CLASS I ON I.OID = X.INDEXRELID JOIN PG_ATTRIBUTE A ON I.OID = A.ATTRELID " \
			f"WHERE C.RELKIND IN ('r', 'm', 'p')  AND I.RELKIND IN ('i', 'I') " \
			f"AND UPPER(C.RELNAME) = UPPER('{table_name}') AND A.ATTNUM > 0 " \
			f"ORDER BY NON_UNIQUE, INDEX_NAME, COLUMN_NAME"

	def build_literal(self, tables: List[Table], a_literal: Literal, build_plain_value: Callable[[Any], Any] = None):
		return build_literal(tables, a_literal, build_plain_value)



from __future__ import annotations

from watchmen_model.system import DataSource, DataSourceType
from .data_source_postgresql import PostgreSQLDataSourceHelper, PostgreSQLDataSourceParams
from .storage_postgresql import StoragePostgreSQL, TopicDataStoragePostgreSQL


# noinspection PyRedeclaration
class Configuration:
	dataSource: DataSource = DataSource(dataSourceType=DataSourceType.POSTGRESQL)
	params: PostgreSQLDataSourceParams = PostgreSQLDataSourceParams()

	def host(self, host: str, port: int = 3306) -> Configuration:
		self.dataSource.host = host
		self.dataSource.port = str(port)
		return self

	def account(self, username: str, password: str) -> Configuration:
		self.dataSource.username = username
		self.dataSource.password = password
		return self

	def url(self, url: str) -> Configuration:
		"""
		url should include account information, like:
		1. postgresql://username:password@host:port/dbname
		2. postgresql+psycopg2://username:password@host:port/dbname
		"""
		self.dataSource.url = url
		return self

	def schema(self, schema: str) -> Configuration:
		self.dataSource.name = schema
		return self

	def echo(self, enabled: bool = False) -> Configuration:
		self.params.echo = enabled
		return self

	def build(self) -> StoragePostgreSQL:
		return StoragePostgreSQLConfiguration(self.dataSource, self.params).create_storage()

	def build_topic_data(self) -> TopicDataStoragePostgreSQL:
		return StoragePostgreSQLConfiguration(self.dataSource, self.params).create_topic_data_storage()


class StoragePostgreSQLConfiguration:
	"""
	configuration of PostgreSQL storage
	"""

	def __init__(self, data_source: DataSource, params: PostgreSQLDataSourceParams):
		super().__init__()
		self.helper = PostgreSQLDataSourceHelper(data_source, params)

	def create_storage(self) -> StoragePostgreSQL:
		return self.helper.acquire_storage()

	def create_topic_data_storage(self) -> TopicDataStoragePostgreSQL:
		return self.helper.acquire_topic_data_storage()

	@staticmethod
	def config() -> Configuration:
		return Configuration()



from datetime import date, datetime, time
from decimal import Decimal
from typing import Any, Callable, List, Tuple

from sqlalchemy import and_, case, DECIMAL, func, literal, literal_column, or_, Table, TIME
from sqlalchemy.sql.expression import cast

from watchmen_storage import as_table_name, ColumnNameLiteral, ComputedLiteral, ComputedLiteralOperator, \
	EntityCriteria, EntityCriteriaExpression, EntityCriteriaJoint, EntityCriteriaJointConjunction, \
	EntityCriteriaOperator, EntityCriteriaStatement, Literal, NoCriteriaForUpdateException, \
	UnexpectedStorageException, UnsupportedComputationException, UnsupportedCriteriaException
from watchmen_storage_rds import SQLAlchemyStatement
from watchmen_utilities import ArrayHelper, DateTimeConstants, is_blank, is_not_blank


def to_decimal(value: Any) -> Any:
	return cast(value, DECIMAL)


DATE_FORMAT_MAPPING = {
	'Y': 'YYYY',  # 4 digits year
	'y': 'YY',  # 2 digits year
	'M': 'MM',  # 2 digits month
	'D': 'DD',  # 2 digits day of month
	'h': 'HH24',  # 2 digits hour, 00 - 23
	'H': 'HH',  # 2 digits hour, 01 - 12
	'm': 'MI',  # 2 digits minute
	's': 'SS',  # 2 digits second
	'W': 'FMDay',  # Monday - Sunday
	'w': 'Dy',  # Mon - Sun
	'B': 'FMMonth',  # January - December
	'b': 'Mon',  # Jan - Dec
	'p': 'PM'  # AM/PM
}


def translate_date_format(date_format: str) -> str:
	return ArrayHelper(list(DATE_FORMAT_MAPPING)) \
		.reduce(lambda original, x: original.replace(x, DATE_FORMAT_MAPPING[x]), date_format)


# noinspection DuplicatedCode
def build_literal(tables: List[Table], a_literal: Literal, build_plain_value: Callable[[Any], Any] = None):
	if isinstance(a_literal, ColumnNameLiteral):
		if is_blank(a_literal.entityName):
			# table name is not given
			if len(tables) == 0:
				# in subquery, no table passed-in
				return literal_column(a_literal.columnName)
			elif len(tables) != 1:
				raise UnexpectedStorageException(
					'Available table must be unique when entity name is missed in column name literal.')
			else:
				# noinspection PyPropertyAccess
				return tables[0].c[a_literal.columnName]
		else:
			table_name = as_table_name(a_literal)
			table = ArrayHelper(tables).find(lambda x: x.name == table_name)
			if table is None:
				raise UnexpectedStorageException(f'Entity[{a_literal.entityName}] not found.')
			return table.c[a_literal.columnName]
	elif isinstance(a_literal, ComputedLiteral):
		operator = a_literal.operator
		if operator == ComputedLiteralOperator.ADD:
			return ArrayHelper(a_literal.elements) \
				.map(lambda x: build_literal(tables, x, to_decimal)) \
				.reduce(lambda prev, current: prev + current, None)
		elif operator == ComputedLiteralOperator.SUBTRACT:
			return ArrayHelper(a_literal.elements) \
				.map(lambda x: build_literal(tables, x, to_decimal)) \
				.reduce(lambda prev, current: prev - current, None)
		elif operator == ComputedLiteralOperator.MULTIPLY:
			return ArrayHelper(a_literal.elements) \
				.map(lambda x: build_literal(tables, x, to_decimal)) \
				.reduce(lambda prev, current: prev * current, None)
		elif operator == ComputedLiteralOperator.DIVIDE:
			return ArrayHelper(a_literal.elements) \
				.map(lambda x: build_literal(tables, x, to_decimal)) \
				.reduce(lambda prev, current: prev / current, None)
		elif operator == ComputedLiteralOperator.MODULUS:
			return ArrayHelper(a_literal.elements) \
				.map(lambda x: build_literal(tables, x, to_decimal)) \
				.reduce(lambda prev, current: prev % current, None)
		elif operator == ComputedLiteralOperator.YEAR_OF:
			# year is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.year(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.HALF_YEAR_OF:
			# month is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return case((
				func.month(build_literal(tables, a_literal.elements[0])) <= 6, DateTimeConstants.HALF_YEAR_FIRST.value),
				else_=DateTimeConstants.HALF_YEAR_SECOND.value
			)
		elif operator == ComputedLiteralOperator.QUARTER_OF:
			# quarter is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.quarter(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.MONTH_OF:
			# month is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.month(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.WEEK_OF_YEAR:
			# week is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.week(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.WEEK_OF_MONTH:
			# weekofmonth is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.weekofmonth(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.DAY_OF_MONTH:
			# day is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.day(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.DAY_OF_WEEK:
			# weekday is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.weekday(build_literal(tables, a_literal.elements[0]))
		elif operator == ComputedLiteralOperator.CASE_THEN:
			elements = a_literal.elements
			cases = ArrayHelper(elements).filter(lambda x: isinstance(x, Tuple)) \
				.map(lambda x: (build_criteria_statement(tables, x[0]), build_literal(tables, x[1]))) \
				.to_list()
			anyway = ArrayHelper(elements).find(lambda x: not isinstance(x, Tuple))
			if anyway is None:
				return case(*cases)
			else:
				return case(*cases, else_=build_literal(tables, anyway))
		elif operator == ComputedLiteralOperator.CONCAT:
			return func.concat(*ArrayHelper(a_literal.elements).map(lambda x: build_literal(tables, x)).to_list())
		elif operator == ComputedLiteralOperator.YEAR_DIFF:
			# yeardiff is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.yeardiff(
				build_literal(tables, a_literal.elements[0]), build_literal(tables, a_literal.elements[1]))
		elif operator == ComputedLiteralOperator.MONTH_DIFF:
			# monthdiff is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.monthdiff(
				build_literal(tables, a_literal.elements[0]), build_literal(tables, a_literal.elements[1]))
		elif operator == ComputedLiteralOperator.DAY_DIFF:
			# datediff is a customized function, which can be found in data-scripts folder
			# make sure each topic storage have this function
			return func.datediff(
				build_literal(tables, a_literal.elements[0]), build_literal(tables, a_literal.elements[1]))
		elif operator == ComputedLiteralOperator.MOVE_DATE:
			return func.movedate(
				build_literal(tables, a_literal.elements[0]), a_literal.elements[1])
		elif operator == ComputedLiteralOperator.FORMAT_DATE:
			return func.to_char(
				build_literal(tables, a_literal.elements[0]), translate_date_format(a_literal.elements[1]))
		elif operator == ComputedLiteralOperator.CHAR_LENGTH:
			return func.length(func.coalesce(build_literal(tables, a_literal.elements[0]), ''))
		else:
			raise UnsupportedComputationException(f'Unsupported computation operator[{operator}].')
	elif isinstance(a_literal, datetime):
		return func.to_timestamp(a_literal.strftime('%Y-%m-%d %H:%M:%S'), 'YYYY-MM-DD HH24:MI:SS')
	elif isinstance(a_literal, date):
		return func.to_date(a_literal.strftime('%Y-%m-%d'), 'YYYY-MM-DD')
	elif isinstance(a_literal, time):
		return cast(func.to_timestamp(a_literal.strftime('%H:%M:%S'), 'HH24:MI:SS'), TIME)
	elif build_plain_value is not None:
		return build_plain_value(a_literal)
	else:
		# a value, return itself
		return a_literal


# noinspection DuplicatedCode
def build_criteria_expression(tables: List[Table], expression: EntityCriteriaExpression):
	built_left = build_literal(tables, expression.left)
	op = expression.operator
	if op == EntityCriteriaOperator.IS_EMPTY:
		return or_(built_left.is_(None), built_left == '')
	elif op == EntityCriteriaOperator.IS_NOT_EMPTY:
		return and_(built_left.is_not(None), built_left != '')
	elif op == EntityCriteriaOperator.IS_BLANK:
		return or_(built_left.is_(None), func.trim(built_left) == '')
	elif op == EntityCriteriaOperator.IS_NOT_BLANK:
		return and_(built_left.is_not(None), func.trim(built_left) != '')

	if op == EntityCriteriaOperator.IN or op == EntityCriteriaOperator.NOT_IN:
		if isinstance(expression.right, ColumnNameLiteral):
			raise UnsupportedCriteriaException('In or not-in criteria expression on another column is not supported.')
		elif isinstance(expression.right, ComputedLiteral):
			if expression.right.operator == ComputedLiteralOperator.CASE_THEN:
				# TODO cannot know whether the built literal will returns a list or a value, let it be now.
				built_right = build_literal(tables, expression.right)
			else:
				# any other computation will not lead a list
				built_right = [build_literal(tables, expression.right)]
		elif isinstance(expression.right, str):
			built_right = ArrayHelper(expression.right.strip().split(',')).filter(lambda x: is_not_blank(x)).to_list()
		else:
			built_right = build_literal(tables, expression.right)
			if not isinstance(built_right, list):
				built_right = [built_right]
		if op == EntityCriteriaOperator.IN:
			if isinstance(built_left, (bool, str, int, float, Decimal, date, time)):
				return literal(built_left).in_(built_right)
			else:
				return built_left.in_(built_right)
		elif op == EntityCriteriaOperator.NOT_IN:
			if isinstance(built_left, (bool, str, int, float, Decimal, date, time)):
				return literal(built_left).not_in(built_right)
			else:
				return built_left.not_in(built_right)

	built_right = build_literal(tables, expression.right)
	if op == EntityCriteriaOperator.EQUALS:
		return built_left == built_right
	elif op == EntityCriteriaOperator.NOT_EQUALS:
		return built_left != built_right
	elif op == EntityCriteriaOperator.LESS_THAN:
		return built_left < built_right
	elif op == EntityCriteriaOperator.LESS_THAN_OR_EQUALS:
		return built_left <= built_right
	elif op == EntityCriteriaOperator.GREATER_THAN:
		return built_left > built_right
	elif op == EntityCriteriaOperator.GREATER_THAN_OR_EQUALS:
		return built_left >= built_right
	elif op == EntityCriteriaOperator.LIKE:
		return built_left.ilike(f'%{built_right}%')
	elif op == EntityCriteriaOperator.NOT_LIKE:
		return built_left.not_ilike(f'%{built_right}%')
	else:
		raise UnsupportedCriteriaException(f'Unsupported criteria expression operator[{op}].')


def build_criteria_joint(tables: List[Table], joint: EntityCriteriaJoint):
	conjunction = joint.conjunction
	if conjunction == EntityCriteriaJointConjunction.AND:
		return and_(*ArrayHelper(joint.children).map(lambda x: build_criteria_statement(tables, x)).to_list())
	elif conjunction == EntityCriteriaJointConjunction.OR:
		return or_(*ArrayHelper(joint.children).map(lambda x: build_criteria_statement(tables, x)).to_list())
	else:
		raise UnsupportedCriteriaException(f'Unsupported criteria joint conjunction[{conjunction}].')


def build_criteria_statement(tables: List[Table], statement: EntityCriteriaStatement):
	if isinstance(statement, EntityCriteriaExpression):
		return build_criteria_expression(tables, statement)
	elif isinstance(statement, EntityCriteriaJoint):
		return build_criteria_joint(tables, statement)
	else:
		raise UnsupportedCriteriaException(f'Unsupported criteria[{statement}].')


def build_criteria(tables: List[Table], criteria: EntityCriteria):
	if criteria is None or len(criteria) == 0:
		return None

	if len(criteria) == 1:
		return build_criteria_statement(tables, criteria[0])
	else:
		return build_criteria_statement(tables, EntityCriteriaJoint(children=criteria))


def build_criteria_for_statement(
		tables: List[Table], statement: SQLAlchemyStatement, criteria: EntityCriteria,
		raise_exception_on_missed: bool = False
) -> SQLAlchemyStatement:
	where = build_criteria(tables, criteria)
	if where is not None:
		return statement.where(where)
	elif raise_exception_on_missed:
		raise NoCriteriaForUpdateException(f'No criteria found from[{criteria}].')
	else:
		return statement
